__author__ = "Benoît Vinot, Victor Gouin, Florent Cadoux, Sébastien Vallet"
__copyright__ = "Roseau Technologies 2018-2022"
__credits__ = "Roseau Technologies"
__license__ = "Proprietary"
__maintainer__ = "Sébastien Vallet"
__email__ = "sebastien.vallet@roseautechnologies.com"
__status__ = "In development"
__url__ = "https://github.com/RoseauTechnologies/Roseau_Load_Flow/"
